import dataclasses
import datetime
import functools
import json
import logging
import pathlib
import sys
import typing
from urllib.parse import urlparse

import click
import rich
import rich.box
from packaging.requirements import Requirement
from packaging.utils import canonicalize_name, parse_wheel_filename
from packaging.version import Version
from rich.table import Table
from rich.text import Text

from fromager import (
    build_environment,
    clickext,
    context,
    hooks,
    metrics,
    overrides,
    progress,
    read,
    server,
    sources,
    wheels,
)

logger = logging.getLogger(__name__)


@dataclasses.dataclass()
@functools.total_ordering
class BuildSequenceEntry:
    name: str
    version: Version
    prebuilt: bool
    download_url: str
    wheel_filename: pathlib.Path
    skipped: bool = False

    @staticmethod
    def dict_factory(x):
        return {
            k: str(v) if isinstance(v, pathlib.Path | Version) else v for (k, v) in x
        }

    def __lt__(self, other):
        if not isinstance(other, BuildSequenceEntry):
            return NotImplemented
        # sort by lower name and version
        return (self.name.lower(), self.version) < (other.name.lower(), other.version)


@click.command()
@click.option(
    "--wheel-server-url",
    default="",
    type=str,
    help="URL for the wheel server for builds",
)
@click.argument("dist_name")
@click.argument("dist_version", type=clickext.PackageVersion())
@click.argument("sdist_server_url")
@click.pass_obj
def build(
    wkctx: context.WorkContext,
    wheel_server_url: str,
    dist_name: str,
    dist_version: Version,
    sdist_server_url: str,
) -> None:
    """Build a single version of a single wheel

    DIST_NAME is the name of a distribution

    DIST_VERSION is the version to process

    SDIST_SERVER_URL is the URL for a PyPI-compatible package index hosting sdists

    1. Downloads the source distribution.

    2. Unpacks it and prepares the source via patching, vendoring rust
       dependencies, etc.

    3. Prepares a build environment with the build dependencies.

    4. Builds the wheel.

    Refer to the 'step' commands for scripting these stages
    separately.

    """
    wkctx.wheel_server_url = wheel_server_url
    server.start_wheel_server(wkctx)
    req = Requirement(f"{dist_name}=={dist_version}")
    source_url, version = sources.resolve_source(
        ctx=wkctx,
        req=req,
        sdist_server_url=sdist_server_url,
    )
    wheel_filename = _build(wkctx, version, req, source_url)
    print(wheel_filename)


build._fromager_show_build_settings = True  # type: ignore


@click.command()
@click.option(
    "-f",
    "--force",
    is_flag=True,
    default=False,
    help="rebuild wheels even if they have already been built",
)
@click.option(
    "-c",
    "--cache-wheel-server-url",
    "cache_wheel_server_url",
    help="url to a wheel server from where fromager can check if it had already built the wheel",
)
@click.argument("build_order_file")
@click.pass_obj
def build_sequence(
    wkctx: context.WorkContext,
    build_order_file: str,
    force: bool,
    cache_wheel_server_url: str | None,
) -> None:
    """Build a sequence of wheels in order

    BUILD_ORDER_FILE is the build-order.json files to build

    SDIST_SERVER_URL is the URL for a PyPI-compatible package index hosting sdists

    Performs the equivalent of the 'build' command for each item in
    the build order file.

    """
    server.start_wheel_server(wkctx)
    wheel_server_urls = [wkctx.wheel_server_url]
    if cache_wheel_server_url:
        # put after local server so we always check local server first
        wheel_server_urls.append(cache_wheel_server_url)

    if force:
        logger.info(f"rebuilding all wheels even if they exist in {wheel_server_urls}")
    else:
        logger.info(
            f"skipping builds for versions of packages available at {wheel_server_urls}"
        )

    entries: list[BuildSequenceEntry] = []

    logger.info("reading build order from %s", build_order_file)
    with read.open_file_or_url(build_order_file) as f:
        for entry in progress.progress(json.load(f)):
            dist_name = entry["dist"]
            resolved_version = Version(entry["version"])
            prebuilt = entry["prebuilt"]
            source_download_url = entry["source_url"]

            req = Requirement(f"{dist_name}=={resolved_version}")

            if not force:
                is_built, wheel_filename = _is_wheel_built(
                    wkctx, dist_name, resolved_version, wheel_server_urls
                )
                if is_built:
                    logger.info(
                        "%s: skipping building wheels for %s==%s since it already exists",
                        dist_name,
                        dist_name,
                        resolved_version,
                    )
                    entries.append(
                        BuildSequenceEntry(
                            dist_name,
                            resolved_version,
                            prebuilt,
                            source_download_url,
                            wheel_filename=wheel_filename,
                            skipped=True,
                        )
                    )
                    continue

            if prebuilt:
                logger.info(
                    "%s: downloading prebuilt wheel %s==%s",
                    dist_name,
                    dist_name,
                    resolved_version,
                )
                wheel_filename = wheels.download_wheel(
                    req=req,
                    wheel_url=source_download_url,
                    output_directory=wkctx.wheels_build,
                )
            else:
                logger.info(
                    "%s: building %s==%s", dist_name, dist_name, resolved_version
                )
                wheel_filename = _build(
                    wkctx, resolved_version, req, source_download_url
                )

            server.update_wheel_mirror(wkctx)
            # After we update the wheel mirror, the built file has
            # moved to a new directory.
            wheel_filename = wkctx.wheels_downloads / wheel_filename.name
            entries.append(
                BuildSequenceEntry(
                    dist_name,
                    resolved_version,
                    prebuilt,
                    source_download_url,
                    wheel_filename=wheel_filename,
                )
            )
            print(wheel_filename)
    metrics.summarize(wkctx, "Building")

    _summary(wkctx, entries)


build_sequence._fromager_show_build_settings = True  # type: ignore


def _summary(ctx: context.WorkContext, entries: list[BuildSequenceEntry]) -> None:
    output: list[typing.Any] = []
    now = datetime.datetime.utcnow().strftime("%Y-%m-%d %H:%M:%SZ")
    output.append(Text(f"Build sequence summary {now}\n"))

    built_entries = [e for e in entries if not e.skipped and not e.prebuilt]
    if built_entries:
        output.append(
            _create_table(
                built_entries,
                title="New builds",
                box=rich.box.MARKDOWN,
                title_justify="left",
            )
        )
    else:
        output.append(Text("No new builds\n"))

    prebuilt_entries = [e for e in entries if e.prebuilt]
    if prebuilt_entries:
        output.append(
            _create_table(
                prebuilt_entries,
                title="Prebuilt wheels",
                box=rich.box.MARKDOWN,
                title_justify="left",
            )
        )
    else:
        output.append(Text("No pre-built wheels\n"))

    skipped_entries = [e for e in entries if e.skipped and not e.prebuilt]
    if skipped_entries:
        output.append(
            _create_table(
                skipped_entries,
                title="Skipped existing builds",
                box=rich.box.MARKDOWN,
                title_justify="left",
            )
        )
    else:
        output.append(Text("No skipped builds\n"))

    console = rich.get_console()
    console.print(*output, sep="\n\n")

    with open(ctx.work_dir / "build-sequence-summary.md", "w", encoding="utf-8") as f:
        console = rich.console.Console(file=f, width=sys.maxsize)
        console.print(*output, sep="\n\n")

    with open(ctx.work_dir / "build-sequence-summary.json", "w", encoding="utf-8") as f:
        json.dump(
            [
                dataclasses.asdict(e, dict_factory=BuildSequenceEntry.dict_factory)
                for e in entries
            ],
            f,
        )


def _create_table(entries: list[BuildSequenceEntry], **table_kwargs) -> Table:
    table = Table(**table_kwargs)
    table.add_column("Name", justify="right", no_wrap=True)
    table.add_column("Version", no_wrap=True)
    table.add_column("Wheel", no_wrap=True)
    table.add_column("Source URL")

    platlib_count = 0

    for info in sorted(entries):
        tags = parse_wheel_filename(info.wheel_filename.name)[3]
        if any(t.platform != "any" or t.abi != "none" for t in tags):
            platlib_count += 1
        source_filename = urlparse(info.download_url).path.rsplit("/", 1)[-1]
        table.add_row(
            info.name,
            str(info.version),
            info.wheel_filename.name,
            # escape Rich markup
            rf"\[{source_filename}]({info.download_url})",
        )

    # summary
    table.add_section()
    table.add_row(
        f"total: {len(entries)}",
        None,
        f"platlib: {platlib_count}",
        None,
    )
    return table


def _build(
    wkctx: context.WorkContext,
    resolved_version: Version,
    req: Requirement,
    source_download_url: str,
) -> pathlib.Path:
    per_wheel_logger = logging.getLogger("")

    module_name = overrides.pkgname_to_override_module(req.name)
    log_filename = module_name + "_current.log"

    wheel_log = wkctx.logs_dir / log_filename

    file_handler = logging.FileHandler(str(wheel_log), mode="w")
    per_wheel_logger.addHandler(file_handler)

    source_filename = sources.download_source(
        ctx=wkctx,
        req=req,
        version=resolved_version,
        download_url=source_download_url,
    )
    logger.debug(
        "%s: saved sdist of version %s from %s to %s",
        req.name,
        resolved_version,
        source_download_url,
        source_filename,
    )

    # Prepare source
    source_root_dir = sources.prepare_source(
        ctx=wkctx, req=req, source_filename=source_filename, version=resolved_version
    )

    # Build environment
    build_environment.prepare_build_environment(
        ctx=wkctx, req=req, sdist_root_dir=source_root_dir
    )
    build_env = build_environment.BuildEnvironment(wkctx, source_root_dir.parent, None)

    # Make a new source distribution, in case we patched the code.
    sdist_filename = sources.build_sdist(
        ctx=wkctx,
        req=req,
        version=resolved_version,
        sdist_root_dir=source_root_dir,
        build_env=build_env,
    )

    # Build
    wheel_filename = wheels.build_wheel(
        ctx=wkctx,
        req=req,
        sdist_root_dir=source_root_dir,
        version=resolved_version,
        build_env=build_env,
    )

    per_wheel_logger.removeHandler(file_handler)
    file_handler.close()

    new_filename = wheel_log.with_name(wheel_filename.stem + ".log")
    wheel_log.rename(new_filename)

    hooks.run_post_build_hooks(
        ctx=wkctx,
        req=req,
        dist_name=canonicalize_name(req.name),
        dist_version=str(resolved_version),
        sdist_filename=sdist_filename,
        wheel_filename=wheel_filename,
    )

    return wheel_filename


def _is_wheel_built(
    wkctx: context.WorkContext,
    dist_name: str,
    resolved_version: Version,
    wheel_server_urls: list[str],
) -> tuple[True, pathlib.Path] | tuple[False, None]:
    req = Requirement(f"{dist_name}=={resolved_version}")

    try:
        logger.info(f"{req.name}: checking if {req} was already built")
        url, _ = wheels.resolve_prebuilt_wheel(
            ctx=wkctx,
            req=req,
            wheel_server_urls=wheel_server_urls,
        )
        pbi = wkctx.package_build_info(req)
        build_tag_from_settings = pbi.build_tag(resolved_version)
        build_tag = build_tag_from_settings if build_tag_from_settings else (0, "")
        wheel_filename = urlparse(url).path.rsplit("/", 1)[-1]
        _, _, build_tag_from_name, _ = parse_wheel_filename(wheel_filename)
        existing_build_tag = build_tag_from_name if build_tag_from_name else (0, "")
        if (
            existing_build_tag[0] > build_tag[0]
            and existing_build_tag[1] == build_tag[1]
        ):
            raise ValueError(
                f"{dist_name}: changelog for version {resolved_version} is inconsistent. Found build tag {existing_build_tag} but expected {build_tag}"
            )
        is_built = existing_build_tag == build_tag
        if is_built and wkctx.wheel_server_url not in url:
            # if the found wheel was on an external server, then download it
            wheels.download_wheel(req, url, wkctx.wheels_downloads)
            server.update_wheel_mirror(wkctx)

        return is_built, pathlib.Path(wheel_filename)
    except Exception:
        logger.info(f"{req.name}: could not locate prebuilt wheel. Will build {req}")
        return False, None



import logging
import typing

import click
from packaging.requirements import Requirement
from packaging.utils import NormalizedName
from packaging.version import Version

from .. import (
    bootstrapper,
    context,
    dependency_graph,
    metrics,
    progress,
    requirements_file,
    resolver,
    server,
    sources,
    wheels,
)
from .graph import find_why

# Map child_name==child_version to list of (parent_name==parent_version, Requirement)
ReverseRequirements = dict[str, list[tuple[str, Requirement]]]

logger = logging.getLogger(__name__)


def _get_requirements_from_args(
    toplevel: typing.Iterable[str],
    req_files: typing.Iterable[str],
) -> list[Requirement]:
    parsed_req: list[str] = []
    parsed_req.extend(toplevel)
    for filename in req_files:
        parsed_req.extend(requirements_file.parse_requirements_file(filename))
    to_build: list[Requirement] = []
    for dep in parsed_req:
        req = Requirement(dep)
        # If we're given a requirements file as input, we might be iterating over a
        # list of requirements with marker expressions that limit their use to
        # specific platforms or python versions. Evaluate the markers to filter out
        # anything we shouldn't build. Only apply the filter to toplevel
        # requirements (items without a why list leading up to them) because other
        # dependencies are already filtered based on their markers in the context of
        # their parent, so they include values like the parent's extras settings.
        if not requirements_file.evaluate_marker(req, req):
            logger.info(
                f"{req.name}: ignoring {requirements_file.RequirementType.TOP_LEVEL} dependency {req} because of its marker expression"
            )
        else:
            to_build.append(req)
    return to_build


@click.command()
@click.option(
    "-r",
    "--requirements-file",
    "requirements_files",
    multiple=True,
    type=str,
    help="pip requirements file",
)
@click.option(
    "-p",
    "--previous-bootstrap-file",
    "previous_bootstrap_file",
    type=str,
    help="graph file produced from a previous bootstrap",
)
@click.option(
    "-c",
    "--cache-wheel-server-url",
    "cache_wheel_server_url",
    help="url to a wheel server from where fromager can download the wheels that it has built before",
)
@click.argument("toplevel", nargs=-1)
@click.pass_obj
def bootstrap(
    wkctx: context.WorkContext,
    requirements_files: list[str],
    previous_bootstrap_file: str | None,
    cache_wheel_server_url: str | None,
    toplevel: list[str],
) -> None:
    """Compute and build the dependencies of a set of requirements recursively

    TOPLEVEL is a requirements specification, including a package name
    and optional version constraints.

    """
    logger.info(f"cache wheel server url: {cache_wheel_server_url}")
    to_build = _get_requirements_from_args(toplevel, requirements_files)
    if not to_build:
        raise RuntimeError(
            "Pass a requirement specificiation or use -r to pass a requirements file"
        )
    logger.info("bootstrapping %r variant of %s", wkctx.variant, to_build)

    if previous_bootstrap_file:
        logger.info("reading previous bootstrap data from %s", previous_bootstrap_file)
        prev_graph = dependency_graph.DependencyGraph.from_file(previous_bootstrap_file)
    else:
        logger.info("no previous bootstrap data")
        prev_graph = None

    pre_built = wkctx.settings.list_pre_built()
    if pre_built:
        logger.info("treating %s as pre-built wheels", sorted(pre_built))

    server.start_wheel_server(wkctx)

    # we need to resolve all the top level dependencies before we start bootstrapping.
    # this is to ensure that if we are using an older bootstrap to resolve packages
    # we are able to upgrade a package anywhere in the dependency tree if it is mentioned
    # in the toplevel without having to fall back to history
    logger.info("resolving top-level dependencies before building")
    for req in to_build:
        pbi = wkctx.package_build_info(req)
        if pbi.pre_built:
            servers = wheels.get_wheel_server_urls(wkctx, req)
            source_url, version = wheels.resolve_prebuilt_wheel(
                ctx=wkctx,
                req=req,
                wheel_server_urls=servers,
            )
        else:
            source_url, version = sources.resolve_source(
                ctx=wkctx,
                req=req,
                sdist_server_url=resolver.PYPI_SERVER_URL,
            )
        logger.info("%s resolves to %s", req, version)
        wkctx.dependency_graph.add_dependency(
            parent_name=None,
            parent_version=None,
            req_type=requirements_file.RequirementType.TOP_LEVEL,
            req=req,
            req_version=version,
            download_url=source_url,
            pre_built=pbi.pre_built,
        )

    with progress.progress_context(total=len(to_build)) as progressbar:
        bt = bootstrapper.Bootstrapper(
            wkctx, progressbar, prev_graph, cache_wheel_server_url
        )

        for req in to_build:
            bt.bootstrap(req, requirements_file.RequirementType.TOP_LEVEL)
            progressbar.update()

    # If we put pre-built wheels in the downloads directory, we should
    # remove them so we can treat that directory as a source of wheels
    # to upload to an index.
    for prebuilt_wheel in wkctx.wheels_prebuilt.glob("*.whl"):
        filename = wkctx.wheels_downloads / prebuilt_wheel.name
        if filename.exists():
            logger.info(
                f"removing prebuilt wheel {prebuilt_wheel.name} from download cache"
            )
            filename.unlink()

    constraints_filename = wkctx.work_dir / "constraints.txt"
    logger.info(f"writing installation dependencies to {constraints_filename}")
    with open(constraints_filename, "w") as f:
        if not write_constraints_file(graph=wkctx.dependency_graph, output=f):
            raise ValueError(
                f"Could not produce a pip compatible constraints file. Please review {constraints_filename} for more details"
            )

    metrics.summarize(wkctx, "Bootstrapping")


def write_constraints_file(
    graph: dependency_graph.DependencyGraph,
    output: typing.TextIO,
) -> bool:
    # Look for potential conflicts by tracking how many different versions of
    # each package are needed.
    conflicts = graph.get_install_dependency_versions()
    ret = True
    conflicting_deps = set()

    # Map for already resolved versions for a given dependency Eg: {"a": "0.4"}
    resolved: dict[NormalizedName, Version] = {}

    # List of unresolved dependencies
    unresolved_dependencies = sorted(conflicts.items())

    dep_name: NormalizedName

    # Loop over dependencies and resolve dependencies with single version first. This will shrink the unresolved_dependencies to begin with.
    for dep_name, nodes in unresolved_dependencies[:]:
        versions = [node.version for node in nodes]
        if len(versions) == 0:
            # This should never happen.
            raise ValueError(f"No versions of {dep_name} supported")

        if len(versions) == 1:
            # This is going to be the situation for most dependencies, where we
            # only have one version.
            resolved[dep_name] = versions[0]
            # Remove from unresolved dependencies list
            unresolved_dependencies.remove((dep_name, nodes))
    multiple_versions = dict(unresolved_dependencies)

    # Below this point we have built multiple versions of the same thing, so
    # we need to try to determine if any one of those versions meets all of
    # the requirements.

    # Flag to see if something is resolved
    resolved_something = True

    # Outer while loop to resolve remaining dependencies with multiple versions
    while unresolved_dependencies and resolved_something:
        resolved_something = False
        # Make copy of the original list and loop over unresolved dependencies
        for dep_name, nodes in unresolved_dependencies[:]:
            # Track which versions can be used by which parent requirement.
            usable_versions: dict[str, list[Version]] = {}
            # Track how many total users of a requirement (by name) there are so we
            # can tell later if any version can be used by all of them.
            user_counter = 0
            # Which parent requirements can use which versions of the dependency we
            # are working on?
            dep_versions = [node.version for node in nodes]
            # Loop over the nodes list
            for node in nodes:
                parent_edges = node.get_incoming_install_edges()
                # Loop over parent_edges list
                for parent_edge in parent_edges:
                    parent_name = parent_edge.destination_node.canonicalized_name
                    # Condition to select the right version.
                    # We check whether parent_name is already in resolved dict and the version associated with that
                    # is not the version of the destination node
                    if (
                        parent_name in resolved
                        and resolved[parent_name]
                        != parent_edge.destination_node.version
                    ):
                        continue
                    # Loop to find the usable versions
                    for matching_version in parent_edge.req.specifier.filter(
                        dep_versions
                    ):
                        usable_versions.setdefault(str(matching_version), []).append(
                            parent_edge.destination_node.version
                        )
                    user_counter += 1

            # Look for one version that can be used by all the parent dependencies
            # and output that if we find it. Otherwise, include a warning and report
            # all versions so a human reading the file can make their own decision
            # about how to resolve the conflict.
            for v, users in usable_versions.items():
                if len(users) == user_counter:
                    version_strs = [str(v) for v in sorted(versions)]
                    logger.debug(
                        "%s: selecting %s from multiple candidates %s",
                        dep_name,
                        v,
                        version_strs,
                    )
                    resolved[dep_name] = Version(v)
                    resolved_something = True
                    unresolved_dependencies.remove((dep_name, nodes))

    # Write resolved versions to constraints file
    for dep_name, resolved_version in sorted(resolved.items()):
        if dep_name in multiple_versions:
            version_strs = [
                str(node.version)
                for node in sorted(multiple_versions[dep_name], key=lambda n: n.version)
            ]
            output.write(
                f"# NOTE: fromager selected {dep_name}=={resolved_version} from: {version_strs}\n"
            )
        output.write(f"{dep_name}=={resolved_version}\n")

    # No single version could be used, so go ahead and print all the
    # versions with a warning message
    for dep_name, nodes in unresolved_dependencies:
        ret = False
        logger.error("%s: no single version meets all requirements", dep_name)
        output.write(f"# ERROR: no single version of {dep_name} met all requirements\n")
        conflicting_deps.add(dep_name)
        for node in sorted(nodes, key=lambda n: n.version):
            output.write(f"{dep_name}=={node.version}\n")

    for dep_name in conflicting_deps:
        logger.error("finding why %s was being used", dep_name)
        for node in graph.get_nodes_by_name(dep_name):
            find_why(graph, node, -1, 1, [])

    return ret


bootstrap._fromager_show_build_settings = True  # type: ignore



import pathlib
from unittest.mock import Mock, patch

import pytest
from packaging.requirements import Requirement
from packaging.version import Version

from fromager import constraints


def test_constraint_is_satisfied_by():
    c = constraints.Constraints({"foo": Requirement("foo<=1.1")})
    assert c.is_satisfied_by("foo", "1.1")
    assert c.is_satisfied_by("foo", Version("1.0"))
    assert c.is_satisfied_by("bar", Version("2.0"))


def test_constraint_canonical_name():
    c = constraints.Constraints({"flash_attn": Requirement("flash_attn<=1.1")})
    assert c.is_satisfied_by("flash_attn", "1.1")
    assert c.is_satisfied_by("flash-attn", "1.1")
    assert c.is_satisfied_by("Flash-ATTN", "1.1")


def test_constraint_not_is_satisfied_by():
    c = constraints.Constraints({"foo": Requirement("foo<=1.1")})
    assert not c.is_satisfied_by("foo", "1.2")
    assert not c.is_satisfied_by("foo", Version("2.0"))


def test_load_empty_constraints_file():
    assert constraints.load(None)._data == {}


def test_load_non_existant_constraints_file(tmp_path: pathlib.Path):
    non_existant_file = tmp_path / "non_existant.txt"
    with pytest.raises(FileNotFoundError):
        constraints.load(non_existant_file)


@patch("fromager.requirements_file.parse_requirements_file")
def test_load_constraints_file(parse_requirements_file: Mock, tmp_path: pathlib.Path):
    constraint_file = tmp_path / "constraint.txt"
    constraint_file.write_text("a\n")
    parse_requirements_file.return_value = ["torch==3.1.0"]
    assert constraints.load(constraint_file)._data == {
        "torch": Requirement("torch==3.1.0")
    }



import pathlib
import textwrap

from fromager.requirements_file import RequirementType, parse_requirements_file


def test_get_requirements_requirements_file(tmp_path: pathlib.Path):
    req_file = tmp_path / "requirements.txt"
    req_file.write_text("c\n")
    requirements = parse_requirements_file(req_file)
    assert requirements == ["c"]


def test_get_requirements_requirements_file_comments(tmp_path: pathlib.Path):
    req_file = tmp_path / "requirements.txt"
    req_file.write_text(
        textwrap.dedent("""
        c
        d # with comment
        # ignore comment

        """),
    )
    requirements = parse_requirements_file(req_file)
    assert requirements == ["c", "d"]


def test_get_requirements_file_with_comments_and_blanks(tmp_path: pathlib.Path):
    req_file = tmp_path / "requirements.txt"
    req_file.write_text("a\n\n# ignore\nb\nc\n")
    requirements = parse_requirements_file(req_file)
    assert requirements == ["a", "b", "c"]


def test_compare_req_type():
    assert RequirementType.BUILD == RequirementType.BUILD_BACKEND
    assert RequirementType.BUILD == RequirementType.BUILD_SDIST
    assert RequirementType.BUILD == RequirementType.BUILD_SYSTEM

    # reverse order
    assert RequirementType.BUILD_SYSTEM == RequirementType.BUILD
    assert RequirementType.BUILD_SDIST == RequirementType.BUILD
    assert RequirementType.BUILD_BACKEND == RequirementType.BUILD

    assert RequirementType.INSTALL != RequirementType.BUILD_BACKEND
    assert RequirementType.INSTALL != RequirementType.BUILD_SYSTEM
    assert RequirementType.INSTALL != RequirementType.BUILD_SDIST
    assert RequirementType.INSTALL != RequirementType.BUILD

    # make sure they equal themselves
    for r in RequirementType:
        assert r == r



"""Vendor Rust crates into an sdist"""

import json
import logging
import os
import pathlib
import typing

import tomlkit
from packaging.requirements import Requirement

from . import dependencies, external_commands

logger = logging.getLogger(__name__)

VENDOR_DIR = "vendor"

CARGO_CONFIG = {
    "source": {
        "crates-io": {"replace-with": "vendored-sources"},
        "vendored-sources": {"directory": VENDOR_DIR},
    },
}

# vendor when project depends on a Rust build system
RUST_BUILD_REQUIRES = frozenset({"setuptools-rust", "maturin"})


def _cargo_vendor(
    req: Requirement,
    manifests: list[pathlib.Path],
    project_dir: pathlib.Path,
) -> typing.Iterable[pathlib.Path]:
    """Run cargo vendor"""
    logger.info(f"{req.name}: updating vendored rust dependencies in {project_dir}")
    args = ["cargo", "vendor", f"--manifest-path={manifests[0]}"]
    for manifest in manifests[1:]:
        args.append(f"--sync={manifest}")
    args.append(os.fspath(project_dir / VENDOR_DIR))
    external_commands.run(args, network_isolation=False)
    return sorted(project_dir.joinpath(VENDOR_DIR).iterdir())


def _cargo_shrink(crate_dir: pathlib.Path) -> None:
    """Remove pre-compiled archive and lib files.

    They are only used on Windows, macOS, and iOS. This makes the vendor
    bundle up to 60% smaller. If we ever need these files, then we have to
    figure out how to compile them from sources, too.
    """
    removed_files: list[str] = []
    for pattern in ["**/*.a", "**/*.lib"]:
        for filename in crate_dir.glob(pattern):
            logger.debug("removing '%s'", filename.relative_to(crate_dir.parent))
            filename.unlink()
            filename.touch()  # create empty file
            removed_files.append(str(filename.relative_to(crate_dir)))

    # update checksums
    checksum_file = crate_dir.joinpath(".cargo-checksum.json")
    if removed_files and checksum_file.is_file():
        with checksum_file.open("r", encoding="utf-8") as f:
            checksums = json.load(f)
        for rf in removed_files:
            checksums["files"].pop(rf, None)
        with checksum_file.open("w", encoding="utf-8") as f:
            json.dump(checksums, f)


def _cargo_config(project_dir: pathlib.Path) -> None:
    """create .cargo/config.toml"""
    dotcargo = project_dir / ".cargo"
    config_toml = dotcargo / "config.toml"
    try:
        with open(config_toml, "r", encoding="utf-8") as f:
            cfg = tomlkit.load(f)
        logger.debug("extending existing '.cargo/config.toml'")
    except FileNotFoundError:
        logger.debug("creating new '.cargo/config.toml'")
        dotcargo.mkdir(exist_ok=True)
        cfg = tomlkit.parse("")

    cfg.update(CARGO_CONFIG)

    with open(config_toml, "w", encoding="utf-8") as f:
        tomlkit.dump(cfg, f)


def _should_vendor_rust(req: Requirement, project_dir: pathlib.Path) -> bool:
    """Detect if project has build requirement on Rust

    Detects setuptools-rust and maturin.
    """
    pyproject_toml = dependencies.get_pyproject_contents(project_dir)
    if not pyproject_toml:
        logger.debug(f"{req.name}: has no pyproject.toml")
        return False

    build_backend = dependencies.get_build_backend(pyproject_toml)

    for reqstring in build_backend["requires"]:
        req = Requirement(reqstring)
        if req.name in RUST_BUILD_REQUIRES:
            logger.debug(
                f"{req.name}: build-system requires {req.name}, vendoring crates"
            )
            return True

    logger.debug(f"{req.name}: no Rust build plugin detected")
    return False


def vendor_rust(
    req: Requirement, project_dir: pathlib.Path, *, shrink_vendored: bool = True
) -> bool:
    """Vendor Rust crates into a source directory

    Returns ``True`` if the project has a build dependency on
    ``setuptools-rust`` or ``maturin``, and has a ``Cargo.toml``, otherwise
    ``False``.
    """
    if not _should_vendor_rust(req, project_dir):
        return False

    # check for Cargo.toml
    manifests = list(project_dir.glob("**/Cargo.toml"))
    if not manifests:
        logger.debug(f"{req.name}: has no Cargo.toml files")
        return False

    the_manifests = sorted(str(d.relative_to(project_dir)) for d in manifests)
    logger.debug(f"{req.name}: {project_dir} has cargo manifests: {the_manifests}")

    # fetch and vendor Rust crates
    vendored = _cargo_vendor(req, manifests, project_dir)
    logger.debug(f"{req.name}: vendored crates: {sorted(d.name for d in vendored)}")

    # remove unnecessary pre-compiled files for Windows, macOS, and iOS.
    if shrink_vendored:
        for crate_dir in vendored:
            _cargo_shrink(crate_dir)
    # update or create .cargo/config.toml
    _cargo_config(project_dir)

    return True



import pathlib
from unittest.mock import patch

import pytest
import wheel.cli  # type: ignore

from fromager import wheels


@patch("fromager.sources.download_url")
def test_invalid_wheel_file_exception(mock_download_url, tmp_path: pathlib.Path):
    mock_download_url.return_value = pathlib.Path(tmp_path / "test" / "fake_wheel.txt")
    fake_url = "https://www.thisisafakeurl.com"
    fake_dir = tmp_path / "test"
    fake_dir.mkdir()
    text_file = fake_dir / "fake_wheel.txt"
    text_file.write_text("This is a test file")
    with pytest.raises(wheel.cli.WheelError):
        wheels._download_wheel_check(fake_dir, fake_url)



from __future__ import annotations

import json
import logging
import operator
import pathlib
import shutil
import typing
import zipfile
from urllib.parse import urlparse

from packaging.requirements import Requirement
from packaging.utils import NormalizedName, canonicalize_name
from packaging.version import Version

from . import (
    build_environment,
    dependencies,
    finders,
    progress,
    resolver,
    server,
    sources,
    wheels,
)
from .dependency_graph import DependencyGraph
from .requirements_file import RequirementType, SourceType

if typing.TYPE_CHECKING:
    from . import context

logger = logging.getLogger(__name__)


class Bootstrapper:
    def __init__(
        self,
        ctx: context.WorkContext,
        progressbar: progress.Progressbar | None = None,
        prev_graph: DependencyGraph | None = None,
        cache_wheel_server_url: str | None = None,
    ) -> None:
        self.ctx = ctx
        self.progressbar = progressbar or progress.Progressbar(None)
        self.prev_graph = prev_graph
        self.cache_wheel_server_url = cache_wheel_server_url
        self.why: list[tuple[RequirementType, Requirement, Version]] = []
        # Push items onto the stack as we start to resolve their
        # dependencies so at the end we have a list of items that need to
        # be built in order.
        self._build_stack: list[typing.Any] = []
        self._build_requirements: set[tuple[NormalizedName, str]] = set()

        # Track requirements we've seen before so we don't resolve the
        # same dependencies over and over and so we can break cycles in
        # the dependency list. The key is the requirements spec, rather
        # than the package, in case we do have multiple rules for the same
        # package.
        self._seen_requirements: set[tuple[NormalizedName, tuple[str, ...], str]] = (
            set()
        )

        self._build_order_filename = self.ctx.work_dir / "build-order.json"

    def bootstrap(self, req: Requirement, req_type: RequirementType) -> Version:
        constraint = self.ctx.constraints.get_constraint(req.name)
        if constraint:
            logger.info(
                f"{req.name}: incoming requirement {req} matches constraint {constraint}. Will apply both."
            )

        pbi = self.ctx.package_build_info(req)
        if pbi.pre_built:
            wheel_url, resolved_version = self._resolve_prebuilt_with_history(
                req=req,
                req_type=req_type,
            )
            source_url = wheel_url
        else:
            source_url, resolved_version = self._resolve_source_with_history(
                req=req,
                req_type=req_type,
            )

        self._add_to_graph(req, req_type, resolved_version, source_url)

        # Avoid cyclic dependencies and redundant processing.
        if self._has_been_seen(req, resolved_version):
            logger.debug(
                f"{req.name}: redundant {req_type} dependency {req} ({resolved_version}) for {self._explain}"
            )
            return resolved_version
        self._mark_as_seen(req, resolved_version)

        logger.info(
            f"{req.name}: new {req_type} dependency {req} resolves to {resolved_version}"
        )

        # Build the dependency chain up to the point of this new
        # requirement using a new list so we can avoid modifying the list
        # we're given.
        self.why.append((req_type, req, resolved_version))

        # for cleanup
        build_env = None
        sdist_root_dir = None
        if pbi.pre_built:
            wheel_filename, unpack_dir = self._download_prebuilt(
                req=req,
                req_type=req_type,
                resolved_version=resolved_version,
                wheel_url=source_url,
            )
            # Remember that this is a prebuilt wheel, and where we got it.
            source_url_type = str(SourceType.PREBUILT)
        else:
            unpacked_cached_wheel, cached_wheel_filename = (
                self._download_wheel_from_cache(req, resolved_version)
            )
            source_url_type = sources.get_source_type(self.ctx, req)

            if not unpacked_cached_wheel:
                source_filename = sources.download_source(
                    ctx=self.ctx,
                    req=req,
                    version=resolved_version,
                    download_url=source_url,
                )
                sdist_root_dir = sources.prepare_source(
                    ctx=self.ctx,
                    req=req,
                    source_filename=source_filename,
                    version=resolved_version,
                )
                unpack_dir = sdist_root_dir.parent
            else:
                sdist_root_dir = unpacked_cached_wheel
                unpack_dir = unpacked_cached_wheel.parent

            # need to call this function irrespective of whether we had the wheel cached
            # so that the build dependencies can be bootstrapped
            build_dependencies = self._prepare_build_dependencies(req, sdist_root_dir)

            # skip building even if it is a non-fromager built wheel
            if not cached_wheel_filename:
                wheel_filename, build_env = self._build(
                    req, resolved_version, sdist_root_dir, build_dependencies
                )
            else:
                wheel_filename = cached_wheel_filename

        self._add_to_build_order(
            req=req,
            version=resolved_version,
            source_url=source_url,
            source_url_type=source_url_type,
            prebuilt=pbi.pre_built,
            constraint=constraint,
        )

        # Process installation dependencies for all wheels.
        install_dependencies = dependencies.get_install_dependencies_of_wheel(
            req, wheel_filename, unpack_dir
        )
        self.progressbar.update_total(len(install_dependencies))
        for dep in self._sort_requirements(install_dependencies):
            try:
                self.bootstrap(dep, RequirementType.INSTALL)
            except Exception as err:
                raise ValueError(f"could not handle {self._explain}") from err
            self.progressbar.update()

        # we are done processing this req, so lets remove it from the why chain
        self.why.pop()
        self._cleanup(req, sdist_root_dir, build_env)
        return resolved_version

    @property
    def _explain(self) -> str:
        """Return message formatting current version of why stack."""
        return " for ".join(
            f"{req_type} dependency {req} ({resolved_version})"
            for req_type, req, resolved_version in reversed(self.why)
        )

    def _build(
        self,
        req: Requirement,
        resolved_version: Version,
        sdist_root_dir: pathlib.Path,
        build_dependencies: set[Requirement],
    ) -> tuple[pathlib.Path, build_environment.BuildEnvironment]:
        build_env = build_environment.BuildEnvironment(
            self.ctx,
            sdist_root_dir.parent,
            build_dependencies,
        )
        try:
            find_sdist_result = finders.find_sdist(
                self.ctx, self.ctx.sdists_builds, req, str(resolved_version)
            )
            if not find_sdist_result:
                sources.build_sdist(
                    ctx=self.ctx,
                    req=req,
                    version=resolved_version,
                    sdist_root_dir=sdist_root_dir,
                    build_env=build_env,
                )
            else:
                logger.info(
                    f"{req.name} have sdist version {resolved_version}: {find_sdist_result}"
                )
        except Exception as err:
            logger.warning(f"{req.name}: failed to build source distribution: {err}")

        logger.info(f"{req.name}: starting build of {self._explain}")
        built_filename = wheels.build_wheel(
            ctx=self.ctx,
            req=req,
            sdist_root_dir=sdist_root_dir,
            version=resolved_version,
            build_env=build_env,
        )
        server.update_wheel_mirror(self.ctx)
        # When we update the mirror, the built file moves to the
        # downloads directory.
        wheel_filename = self.ctx.wheels_downloads / built_filename.name
        logger.info(
            f"{req.name}: built wheel for version {resolved_version}: {wheel_filename}"
        )
        return wheel_filename, build_env

    def _prepare_build_dependencies(
        self, req: Requirement, sdist_root_dir: pathlib.Path
    ) -> set[Requirement]:
        build_system_dependencies = dependencies.get_build_system_dependencies(
            ctx=self.ctx, req=req, sdist_root_dir=sdist_root_dir
        )
        self._handle_build_requirements(
            RequirementType.BUILD_SYSTEM,
            build_system_dependencies,
        )

        build_backend_dependencies = dependencies.get_build_backend_dependencies(
            ctx=self.ctx, req=req, sdist_root_dir=sdist_root_dir
        )
        self._handle_build_requirements(
            RequirementType.BUILD_BACKEND,
            build_backend_dependencies,
        )

        build_sdist_dependencies = dependencies.get_build_sdist_dependencies(
            ctx=self.ctx, req=req, sdist_root_dir=sdist_root_dir
        )
        self._handle_build_requirements(
            RequirementType.BUILD_SDIST,
            build_sdist_dependencies,
        )

        return (
            build_system_dependencies
            | build_backend_dependencies
            | build_sdist_dependencies
        )

    def _handle_build_requirements(
        self, build_type: RequirementType, build_dependencies: set[Requirement]
    ) -> None:
        self.progressbar.update_total(len(build_dependencies))

        for dep in self._sort_requirements(build_dependencies):
            try:
                resolved = self.bootstrap(req=dep, req_type=build_type)
            except Exception as err:
                raise ValueError(f"could not handle {self._explain}") from err
            # We may need these dependencies installed in order to run build hooks
            # Example: frozenlist build-system.requires includes expandvars because
            # it is used by the packaging/pep517_backend/ build backend
            build_environment.maybe_install(self.ctx, dep, build_type, str(resolved))
            self.progressbar.update()

    def _download_prebuilt(
        self,
        req: Requirement,
        req_type: RequirementType,
        resolved_version: Version,
        wheel_url: str,
    ) -> tuple[pathlib.Path, pathlib.Path]:
        logger.info(f"{req.name}: {req_type} requirement {req} uses a pre-built wheel")

        wheel_filename = wheels.download_wheel(req, wheel_url, self.ctx.wheels_prebuilt)

        # Add the wheel to the mirror so it is available to anything
        # that needs to install it. We leave a copy in the prebuilt
        # directory to make it easy to remove the wheel from the
        # downloads directory before uploading to a proper package
        # index.
        dest_name = self.ctx.wheels_downloads / wheel_filename.name
        if not dest_name.exists():
            logger.info(f"{req.name}: updating temporary mirror with pre-built wheel")
            shutil.copy(wheel_filename, dest_name)
            server.update_wheel_mirror(self.ctx)
        unpack_dir = self._create_unpack_dir(req, resolved_version)
        return (wheel_filename, unpack_dir)

    def _download_wheel_from_cache(
        self, req: Requirement, resolved_version: Version
    ) -> tuple[pathlib.Path | None, pathlib.Path | None]:
        if not self.cache_wheel_server_url:
            return None, None
        logger.info(
            f"{req.name}: checking if wheel was already uploaded to {self.cache_wheel_server_url}"
        )
        try:
            wheel_url, _ = resolver.resolve(
                ctx=self.ctx,
                req=Requirement(f"{req.name}=={resolved_version}"),
                sdist_server_url=self.cache_wheel_server_url,
                include_sdists=False,
                include_wheels=True,
            )
            wheelfile_name = pathlib.Path(urlparse(wheel_url).path)
            pbi = self.ctx.package_build_info(req)
            expected_build_tag = pbi.build_tag(resolved_version)
            dist_name, dist_version, build_tag, _ = wheels.extract_info_from_wheel_file(
                req, wheelfile_name
            )
            if expected_build_tag and expected_build_tag != build_tag:
                logger.info(
                    f"{req.name}: found wheel for {resolved_version} in cache but build tag does not match. Got {build_tag} but expected {expected_build_tag}"
                )
                return None, None

            cached_wheel = wheels.download_wheel(
                req=req, wheel_url=wheel_url, output_directory=self.ctx.wheels_downloads
            )
            server.update_wheel_mirror(self.ctx)
            logger.info(f"{req.name}: found built wheel on cache server")
            unpack_dir = self._create_unpack_dir(req, resolved_version)
            dist_filename = f"{dist_name}-{dist_version}"
            metadata_dir = pathlib.Path(f"{dist_filename}.dist-info")
            try:
                archive = zipfile.ZipFile(cached_wheel)
                for filename in [
                    dependencies.BUILD_BACKEND_REQ_FILE_NAME,
                    dependencies.BUILD_SDIST_REQ_FILE_NAME,
                    dependencies.BUILD_SYSTEM_REQ_FILE_NAME,
                ]:
                    zipinfo = archive.getinfo(
                        str(
                            metadata_dir
                            / f"{wheels.FROMAGER_BUILD_REQ_PREFIX}-{filename}"
                        )
                    )
                    zipinfo.filename = filename
                    archive.extract(zipinfo, unpack_dir)

                logger.info(f"{req.name}: extracted build requirements from wheel")
                return unpack_dir / metadata_dir, cached_wheel
            except Exception:
                # implies that the wheel server hosted non-fromager built wheels
                logger.info(
                    f"{req.name}: could not extract build requirements from wheel"
                )
                shutil.rmtree(unpack_dir)
                return None, cached_wheel
        except Exception:
            logger.info(
                f"{req.name}: did not find wheel for {resolved_version} in {self.cache_wheel_server_url}"
            )
            return None, None

    def _resolve_source_with_history(
        self,
        req: Requirement,
        req_type: RequirementType,
    ) -> tuple[str, Version]:
        cached_resolution = self._resolve_from_graph(
            req=req,
            req_type=req_type,
            pre_built=False,
        )
        if cached_resolution:
            source_url, resolved_version = cached_resolution
            logger.debug(
                f"{req.name}: resolved from previous bootstrap to {resolved_version}"
            )
        else:
            source_url, resolved_version = sources.resolve_source(
                ctx=self.ctx,
                req=req,
                sdist_server_url=resolver.PYPI_SERVER_URL,
                req_type=req_type,
            )
        return (source_url, resolved_version)

    def _resolve_prebuilt_with_history(
        self,
        req: Requirement,
        req_type: RequirementType,
    ) -> tuple[str, Version]:
        cached_resolution = self._resolve_from_graph(
            req=req,
            req_type=req_type,
            pre_built=True,
        )

        if cached_resolution:
            wheel_url, resolved_version = cached_resolution
            logger.debug(
                f"{req.name}: resolved from previous bootstrap to {resolved_version}"
            )
        else:
            servers = wheels.get_wheel_server_urls(self.ctx, req)
            wheel_url, resolved_version = wheels.resolve_prebuilt_wheel(
                ctx=self.ctx, req=req, wheel_server_urls=servers, req_type=req_type
            )
        return (wheel_url, resolved_version)

    def _resolve_from_graph(
        self,
        req: Requirement,
        req_type: RequirementType,
        pre_built: bool,
    ) -> tuple[str, Version] | None:
        _, parent_req, _ = self.why[-1] if self.why else (None, None, None)

        # we have already resolved top level reqs before bootstrapping
        # so they should already be in the root node
        if req_type == RequirementType.TOP_LEVEL:
            for edge in self.ctx.dependency_graph.get_root_node().get_outgoing_edges(
                req.name, RequirementType.TOP_LEVEL
            ):
                if edge.req == req:
                    return (
                        edge.destination_node.download_url,
                        edge.destination_node.version,
                    )
            # this should never happen since we already resolved top level reqs and their
            # resolution should be in the root nodes
            raise ValueError(
                f"{req.name}: {req} appears as a toplevel requirement but it's resolution does not exist in the root node of the graph"
            )

        if not self.prev_graph:
            return None

        seen_version: set[str] = set()

        # first perform resolution using the top level reqs before looking at history
        possible_versions_in_top_level: list[tuple[str, Version]] = []
        for (
            top_level_edge
        ) in self.ctx.dependency_graph.get_root_node().get_outgoing_edges(
            req.name, RequirementType.TOP_LEVEL
        ):
            possible_versions_in_top_level.append(
                (
                    top_level_edge.destination_node.download_url,
                    top_level_edge.destination_node.version,
                )
            )
            seen_version.add(str(top_level_edge.destination_node.version))

        resolver_result = self._resolve_from_version_source(
            possible_versions_in_top_level, req
        )
        if resolver_result:
            return resolver_result

        # only if there is nothing in top level reqs, resolve using history
        possible_versions_from_graph: list[tuple[str, Version]] = []
        # check all nodes which have the same parent name irrespective of the parent's version
        for parent_node in self.prev_graph.get_nodes_by_name(
            parent_req.name if parent_req else None
        ):
            # if the edge matches the current req and type then it is a possible candidate
            # filtering on type might not be necessary, but we are being safe here. This will
            # for sure ensure that bootstrap takes the same route as it did in the previous one.
            # If we don't filter by type then it might pick up a different version from a different
            # type that should have appeared much later in the resolution process.
            for edge in parent_node.get_outgoing_edges(req.name, req_type):
                if (
                    edge.destination_node.pre_built == pre_built
                    and str(edge.destination_node.version) not in seen_version
                ):
                    possible_versions_from_graph.append(
                        (
                            edge.destination_node.download_url,
                            edge.destination_node.version,
                        )
                    )
                    seen_version.add(str(edge.destination_node.version))

        return self._resolve_from_version_source(possible_versions_from_graph, req)

    def _resolve_from_version_source(
        self,
        version_source: list[tuple[str, Version]],
        req: Requirement,
    ) -> tuple[str, Version] | None:
        if not version_source:
            return None
        try:
            # no need to pass req type to enable caching since we are already using the graph as our cache
            provider = resolver.GenericProvider(
                version_source=lambda x, y, z: version_source,
                constraints=self.ctx.constraints,
            )
            return resolver.resolve_from_provider(provider, req)
        except Exception as err:
            logger.debug(
                f"{req.name}: could not resolve {req} from {version_source}: {err}"
            )
            return None

    def _create_unpack_dir(self, req: Requirement, resolved_version: Version):
        unpack_dir = self.ctx.work_dir / f"{req.name}-{resolved_version}"
        if not unpack_dir.exists():
            unpack_dir.mkdir()
        return unpack_dir

    def _cleanup(
        self,
        req: Requirement,
        sdist_root_dir: pathlib.Path | None,
        build_env: build_environment.BuildEnvironment | None,
    ) -> None:
        if not self.ctx.cleanup:
            return

        # Cleanup the source tree and build environment, leaving any other
        # artifacts that were created.
        if sdist_root_dir and sdist_root_dir.exists():
            logger.debug(f"{req.name}: cleaning up source tree {sdist_root_dir}")
            shutil.rmtree(sdist_root_dir)
            logger.debug(f"{req.name}: cleaned up source tree {sdist_root_dir}")
        if build_env:
            logger.debug(f"{req.name}: cleaning up build environment {build_env.path}")
            shutil.rmtree(build_env.path)
            logger.debug(f"{req.name}: cleaned up build environment {build_env.path}")

    def _add_to_graph(
        self,
        req: Requirement,
        req_type: RequirementType,
        req_version: Version,
        download_url: str,
    ) -> None:
        if req_type == RequirementType.TOP_LEVEL:
            return

        _, parent_req, parent_version = self.why[-1] if self.why else (None, None, None)
        pbi = self.ctx.package_build_info(req)
        # Update the dependency graph after we determine that this requirement is
        # useful but before we determine if it is redundant so that we capture all
        # edges to use for building a valid constraints file.
        self.ctx.dependency_graph.add_dependency(
            parent_name=canonicalize_name(parent_req.name) if parent_req else None,
            parent_version=parent_version,
            req_type=req_type,
            req=req,
            req_version=req_version,
            download_url=download_url,
            pre_built=pbi.pre_built,
        )
        self.ctx.write_to_graph_to_file()

    def _sort_requirements(
        self,
        requirements: typing.Iterable[Requirement],
    ) -> typing.Iterable[Requirement]:
        return sorted(requirements, key=operator.attrgetter("name"))

    def _resolved_key(
        self, req: Requirement, version: Version
    ) -> tuple[NormalizedName, tuple[str, ...], str]:
        return (canonicalize_name(req.name), tuple(sorted(req.extras)), str(version))

    def _mark_as_seen(self, req: Requirement, version: Version) -> None:
        key = self._resolved_key(req, version)
        logger.debug(f"{req.name}: remembering seen sdist {key}")
        self._seen_requirements.add(key)

    def _has_been_seen(self, req: Requirement, version: Version) -> bool:
        return self._resolved_key(req, version) in self._seen_requirements

    def _add_to_build_order(
        self,
        req: Requirement,
        version: Version,
        source_url: str,
        source_url_type: str,
        prebuilt: bool = False,
        constraint: Requirement | None = None,
    ) -> None:
        # We only care if this version of this package has been built,
        # and don't want to trigger building it twice. The "extras"
        # value, included in the _resolved_key() output, can confuse
        # that so we ignore itand build our own key using just the
        # name and version.
        key = (canonicalize_name(req.name), str(version))
        if key in self._build_requirements:
            return
        logger.info(f"{req.name}: adding {key} to build order")
        self._build_requirements.add(key)
        info = {
            "req": str(req),
            "constraint": str(constraint) if constraint else "",
            "dist": canonicalize_name(req.name),
            "version": str(version),
            "prebuilt": prebuilt,
            "source_url": source_url,
            "source_url_type": source_url_type,
        }
        self._build_stack.append(info)
        with open(self._build_order_filename, "w") as f:
            # Set default=str because the why value includes
            # Requirement and Version instances that can't be
            # converted to JSON without help.
            json.dump(self._build_stack, f, indent=2, default=str)



import logging
import pathlib
import typing

from packaging.requirements import Requirement
from packaging.utils import canonicalize_name
from packaging.version import Version

from . import requirements_file

logger = logging.getLogger(__name__)


class Constraints:
    def __init__(self, data: dict[str, Requirement]):
        self._data = {canonicalize_name(n): v for n, v in data.items()}

    def get_constraint(self, name: str) -> Requirement | None:
        return self._data.get(canonicalize_name(name))

    def allow_prerelease(self, pkg_name: str) -> bool:
        constraint = self.get_constraint(pkg_name)
        if constraint:
            return bool(constraint.specifier.prereleases)
        return False

    def is_satisfied_by(self, pkg_name: str, version: Version) -> bool:
        constraint = self.get_constraint(pkg_name)
        if constraint:
            return constraint.specifier.contains(version, prereleases=True)
        return True


def _parse(content: typing.Iterable[str]) -> Constraints:
    constraints = {}
    for line in content:
        req = Requirement(line)
        if requirements_file.evaluate_marker(req, req):
            constraints[req.name] = req
    return Constraints(constraints)


def load(constraints_file: str | pathlib.Path | None) -> Constraints:
    if not constraints_file:
        return Constraints({})
    logger.info("loading constraints from %s", constraints_file)
    parsed_req_file = requirements_file.parse_requirements_file(constraints_file)
    return _parse(parsed_req_file)



import io
import pathlib
import textwrap

from packaging.requirements import Requirement

from fromager import dependency_graph
from fromager.commands import bootstrap


def test_get_requirements_single_arg():
    requirements = bootstrap._get_requirements_from_args(["a"], [])
    assert [Requirement("a")] == requirements


def test_get_requirements_multiple_args():
    requirements = bootstrap._get_requirements_from_args(["a", "b"], [])
    assert [Requirement("a"), Requirement("b")] == requirements


def test_get_requirements_args_and_file(tmp_path: pathlib.Path):
    requirements_file = tmp_path / "requirements.txt"
    requirements_file.write_text("c\n")
    requirements = bootstrap._get_requirements_from_args(
        ["a", "b"], [str(requirements_file)]
    )
    assert [
        Requirement("a"),
        Requirement("b"),
        Requirement("c"),
    ] == requirements


def test_ignore_based_on_marker():
    requirements = bootstrap._get_requirements_from_args(
        ['foo; python_version<"3.9"'], []
    )
    assert [] == requirements


def test_write_constraints_file_simple():
    buffer = io.StringIO()
    raw_graph = {
        "": {
            "download_url": "",
            "pre_built": False,
            "version": "0",
            "canonicalized_name": "",
            "edges": [{"key": "a==1.0", "req_type": "install", "req": "a"}],
        },
        "a==1.0": {
            "download_url": "url for a",
            "pre_built": False,
            "version": "1.0",
            "canonicalized_name": "a",
            "edges": [
                {"key": "b==2.0", "req_type": "install", "req": "b>=2.0"},
                {"key": "c==3.0", "req_type": "install", "req": "c<4.0"},
            ],
        },
        "b==2.0": {
            "download_url": "url for b",
            "pre_built": False,
            "version": "2.0",
            "canonicalized_name": "b",
            "edges": [],
        },
        "c==3.0": {
            "download_url": "url for c",
            "pre_built": False,
            "version": "3.0",
            "canonicalized_name": "c",
            "edges": [],
        },
    }
    graph = dependency_graph.DependencyGraph.from_dict(raw_graph)
    bootstrap.write_constraints_file(graph, buffer)
    expected = textwrap.dedent("""
        a==1.0
        b==2.0
        c==3.0
        """).lstrip()
    assert expected == buffer.getvalue()


def test_write_constraints_file_resolvable_duplicate():
    buffer = io.StringIO()
    raw_graph = {
        "": {
            "download_url": "",
            "pre_built": False,
            "version": "0",
            "canonicalized_name": "",
            "edges": [{"key": "a==1.0", "req_type": "install", "req": "a"}],
        },
        "a==1.0": {
            "download_url": "url for a",
            "pre_built": False,
            "version": "1.0",
            "canonicalized_name": "a",
            "edges": [
                {"key": "b==2.0", "req_type": "install", "req": "b>=2.0"},
                {"key": "c==3.0", "req_type": "install", "req": "c<4.0"},
            ],
        },
        "b==2.0": {
            "download_url": "url for b",
            "pre_built": False,
            "version": "2.0",
            "canonicalized_name": "b",
            "edges": [{"key": "c==3.1", "req_type": "install", "req": "c>3.0"}],
        },
        "c==3.0": {
            "download_url": "url for c",
            "pre_built": False,
            "version": "3.0",
            "canonicalized_name": "c",
            "edges": [],
        },
        "c==3.1": {
            "download_url": "url for c",
            "pre_built": False,
            "version": "3.1",
            "canonicalized_name": "c",
            "edges": [],
        },
    }
    graph = dependency_graph.DependencyGraph.from_dict(raw_graph)
    bootstrap.write_constraints_file(graph, buffer)
    expected = textwrap.dedent("""
        a==1.0
        b==2.0
        # NOTE: fromager selected c==3.1 from: ['3.0', '3.1']
        c==3.1
        """).lstrip()
    assert expected == buffer.getvalue()


def test_write_constraints_file_unresolvable_duplicate():
    buffer = io.StringIO()
    raw_graph = {
        "": {
            "download_url": "",
            "pre_built": False,
            "version": "0",
            "canonicalized_name": "",
            "edges": [{"key": "a==1.0", "req_type": "install", "req": "a"}],
        },
        "a==1.0": {
            "download_url": "url for a",
            "pre_built": False,
            "version": "1.0",
            "canonicalized_name": "a",
            "edges": [
                {"key": "b==2.0", "req_type": "install", "req": "b>=2.0"},
                {"key": "c==3.0", "req_type": "install", "req": "c==3.0"},
            ],
        },
        "b==2.0": {
            "download_url": "url for b",
            "pre_built": False,
            "version": "2.0",
            "canonicalized_name": "b",
            "edges": [{"key": "c==3.1", "req_type": "install", "req": "c>3.0"}],
        },
        "c==3.0": {
            "download_url": "url for c",
            "pre_built": False,
            "version": "3.0",
            "canonicalized_name": "c",
            "edges": [],
        },
        "c==3.1": {
            "download_url": "url for c",
            "pre_built": False,
            "version": "3.1",
            "canonicalized_name": "c",
            "edges": [],
        },
    }
    graph = dependency_graph.DependencyGraph.from_dict(raw_graph)
    bootstrap.write_constraints_file(graph, buffer)
    expected = textwrap.dedent("""
        a==1.0
        b==2.0
        # ERROR: no single version of c met all requirements
        c==3.0
        c==3.1
        """).lstrip()
    assert expected == buffer.getvalue()


def test_write_constraints_file_duplicates():
    buffer = io.StringIO()
    raw_graph = {
        "": {
            "download_url": "",
            "pre_built": False,
            "version": "0",
            "canonicalized_name": "",
            "edges": [
                {"key": "a==1.0", "req_type": "install", "req": "a"},
                {"key": "d==1.0", "req_type": "install", "req": "d"},
            ],
        },
        "a==1.0": {
            "download_url": "url for a",
            "pre_built": False,
            "version": "1.0",
            "canonicalized_name": "a",
            "edges": [
                {"key": "c==3.0", "req_type": "install", "req": "c<=3.0"},
            ],
        },
        "d==1.0": {
            "download_url": "url for a",
            "pre_built": False,
            "version": "1.0",
            "canonicalized_name": "a",
            "edges": [
                {"key": "c==3.1", "req_type": "install", "req": "c>=3.0"},
            ],
        },
        "c==3.0": {  # transformers 4.46
            "download_url": "url for c",
            "pre_built": False,
            "version": "3.0",
            "canonicalized_name": "c",
            "edges": [{"key": "b==2.0", "req_type": "install", "req": "b<2.1,>=2.0"}],
        },
        "c==3.1": {  # transformers 4.47
            "download_url": "url for c",
            "pre_built": False,
            "version": "3.1",
            "canonicalized_name": "c",
            "edges": [{"key": "b==2.1", "req_type": "install", "req": "b<2.2,>=2.1"}],
        },
        "b==2.0": {  # tokenizer
            "download_url": "url for b",
            "pre_built": False,
            "version": "2.0",
            "canonicalized_name": "b",
            "edges": [],
        },
        "b==2.1": {  # tokenizer
            "download_url": "url for b",
            "pre_built": False,
            "version": "2.1",
            "canonicalized_name": "b",
            "edges": [],
        },
    }
    graph = dependency_graph.DependencyGraph.from_dict(raw_graph)
    assert bootstrap.write_constraints_file(graph, buffer)
    expected = textwrap.dedent("""
        a==1.0
        # NOTE: fromager selected b==2.0 from: ['2.0', '2.1']
        b==2.0
        # NOTE: fromager selected c==3.0 from: ['3.0', '3.1']
        c==3.0
        d==1.0
        """).lstrip()
    assert expected == buffer.getvalue()
